/*------------------------------------------------------------------------------------------------------------
Druckluftschalter
D.Schrder 11/2010

Version 1.1
19.11.2010

 Realisiert mit Tiny25
 
 Interner Oszillator mit 1MHz, 
 (Vorteiler /8 eingeschaltet)
 
 Layout:
             Reset  1    8  +5V
    (PB3)  2     7  (PB2) Magnetventil
    IR-LED   (PB4)  3     6  (PB1) Fototransistor
              GND   4     5  (PB0) Piezo Ausgang
		  

 Input:
  PB1: Emitter des Fototransistors (Collektor an +5V, Emitter mit 47K gegen GND)
  
Output:
  PB0:  Ausgang zum Piezo-Summer. Gibt nach dem Umschalten 2s eine Rechteckfrequenz von 3.7KHz aus (Resonanz des Piezos)
  PB2:  Ausgang zum Treiber des Magnetventils
  PB3:  Ausgang Statusled (High, wenn Lichtschranke nicht unterbrochen) - rote LED
  PB4:  Ausgang zur IR-Sendediode (wird gepulst)
  
Funktion:
  Eine IR-Lichtschranke durchleuchtet einen Druckluftschlauch, in dem sich ein Schwebkrper
  (Alurohr) befindet, der bei Durchfluss angehoben wird und die Lichtschranke unterbricht.
  Die IR-Led wird mit ca. 310 Hz gepulst. Duty-Cycle dabei ca. 10% um Strom zu sparen.
  Ist die Lichtschranke nicht unterbrochen, liefert der Komparator diese Pulse, die gezhlt werden.
  Die Pulse werden ca. 30ms lang gezhlt. 
  Sind es mehr als 5, gilt die Lichtschranke als offen (kein Durchfluss)

  Werden 3 kurze Durchfluesse direkt hintereinander entdeckt, wird das Magnetventil umgeschaltet
  und der Piezo toent fuer ca. 2s.
  Vor und nach der Pulsfolge darf 2s kein Durchfluss sein.

  Nach 7 Minuten ohne Benutzung wird der hohe Druck automatisch abgeschaltet.

*/

#define F_CPU 1000000UL	/* CPU clock in Hertz */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#define U8 uint8_t
#define U16 uint16_t
#define S16 int16_t
#define U32 uint32_t

/* Wie belegt der GCC Register?
R0  ein temporres Register
R1  enthlt immer den Wert 0 
R1..R17, R28, R29
    allgemeine Register, die durch einen Funktionsaufruf nicht verndert bzw wieder auf den 
	ursprnglichen Wert restauriert werden 
R0, R18..R27, R30, R31
    knnen durch Funktionsaufrufe verndert werden, ohne restauriert zu werden 
*/
//Variable an Register binden - Der Prozessor hat zwar genug davon, aber der Compiler nutzt sie nicht.
volatile register U8 pulse         asm("r4");;
register U8 status                 asm("r5");
register U8 durchfluss             asm("r6");
volatile register U8 warten        asm("r7");
register U16 zeit                  asm("r8"); //und R9

//Leider funktioniert "volatile register" nicht, register volatile auch nicht. Der gcc ignoriert das scheinbar.
//Der Compiler liest damit in einer Schleife den Wert vor der Schleife nur 1x aus und packt ihn in r24 
// und testet dann r24 in der Schleife. ndere ich in Interrupt dann r7, ndert das in der Schleife leider nix.
//Also nicht ins Register oder die Schleife in Assembler selber bauen (so habe ich es gemacht)..


//------------------------------------------------------------------------------------------------------------
//ISR(ANA_COMP_vect){ //Steigende Flanke am Komparator
//Spaeshalber mal als nakte Funktion von Hand in Assembler codiert
//siehe auch:    http://www.rn-wissen.de/index.php/Inline-Assembler_in_avr-gcc
void __attribute__ ((naked)) ANA_COMP_vect(void){ 
  asm volatile (
     "push	r0      \n"
     "in	r0, 0x3f  \n"
     "inc r4        \n"  //puls++
     "out	0x3f, r0  \n"
     "pop r0        \n"
     "reti            "
     :
     :
     : "r4" //Sagt dem Compiler, was ich verndert habe

  );
}
//------------------------------------------------------------------------------------------------------------
ISR(TIMER0_COMPA_vect){ //30 mal pro Sekunde die Lichtschranke bewerten
  warten = 0;
}


//------------------------------------------------------------------------------------------------------------
int main(void){
U16 dauerUnbenutzt=0;

U16 i; 

  cli();
  
  DDRB = (1<<PB0)|(1<<PB2)|(1<<PB3)|(1<<PB4); //PB2..4 auf Ausgang
  PORTB = 0; //Alles aus

  for (i=0; i<15; i++){
    PORTB=(1<<PB4)|(1<<PB3);
    _delay_ms(20);
    PORTB=0;
    _delay_ms(20);
  }
  //Probepiepsen
  for (i=0; i<3700; i++){
    PORTB |=1; //Piezo High
    _delay_us(132);
    PORTB &=~1; //Piezo Low
    _delay_us(130);
  }

  //Timer auf PWM, 200Hz Pulsfrequenz, Tastverhltnis auf 10%
  TCCR1 = (1<<CS12)|(1<<CS11); //Teiler /32 = 31.250Hz
  GTCCR = (1<<PWM1B) |(1<<COM1B1);
  OCR1B = 10;
  OCR1C = 100; //312Hz Pulsfrequenz


  TCCR0A = (1<<WGM01); //CTC Modus
  TCCR0B = (1<<CS02)|(1<<CS00); //Vorteiler 1024 (Gibt 976Hz)
  OCR0A  = 31; // Teilt durch 32, gibt  ca. 30Hz
  TIMSK = (1<<OCIE0A); //IRQ alle 33ms

  //Komparator
  ADCSRA = 0; //ADC aus
  ADCSRB = 0; //AIN1 als neg. Eingang (Poti)
  //ACSR = (1<<ACIE) | (1<<ACI) |(1<<ACIS1)|(1<<ACIS0); //IRQ bei steigender Flanke
  ACSR = (1<<ACIE) |(1<<ACBG) | (1<<ACI) |(1<<ACIS1)|(1<<ACIS0); //IRQ bei steigender Flanke

  pulse=0;
  status=0;
  zeit=0;
  durchfluss=0;
  warten=0;
  
  sei();

  #define TOTZEIT_START  75  //ca. 2.5s
  #define MAX_PULSZEIT 10  //ca. 310ms
  #define MAX_PAUSENZEIT 15  //ca. 480ms
  #define MIN_PAUSENZEIT_ENDE 60 //ca. 2s
  #define MIN_PULSZEIT 2
  #define MIN_PAUSENZEIT 2

  while(1){


    //warte, bis 33ms rum
    asm volatile (
        "	and	 r7, r7  \n" 
        " brne	.-4    \n"  // Entspricht: do { } while (warten); // wird vom Timer gelscht
        " inc  r7      \n"  // warten=1;
        );   

     
    //Durchflusserkennung mit Hysterese
    if (0==pulse ){
      durchfluss=1;//Durchfluss unterbricht Lichtschranke
      PORTB &= ~(1<<PB3); //Test-LED aus bei Durchfluss
    }
    else
    if (pulse > 5){
      durchfluss=0;  
      PORTB |= (1<<PB3);  //Test-LED ein bei offener Lichtschranke
    }
    pulse=0;


    //Zeit zhlen
    if (zeit < 30000){
      zeit++;
    }
  
    //Zeit zhlen, die bei hohem Druck ohne Durchfluss vergeht
    if (PORTB & (1<<PB2)){ //Magnetventil ist ein (hoher Druck)
      if (dauerUnbenutzt < 30000){  //1000 Sekunden
        dauerUnbenutzt++;
      }
    }
    if (durchfluss){
      dauerUnbenutzt=0;
    }
  
    //nach 7 Minuten ohne Durchfluss hohen Druck abschalten
    if (dauerUnbenutzt > 30*60*7){  // =12600: 
      PORTB &= ~(1<<PB2); //Magnetventil aus;
    }
  
    //Detektion der Pulsfolge mittels Statusmaschine
    switch(status){
      case 0://Warte, bis mind. 2s kein Durchfluss
        if (durchfluss){
          if (zeit >= TOTZEIT_START){
            status++;
          }
          zeit=0;
        }
        break;
      
      case 1:  //Teste: 1. kurzer Puls 
        if (zeit > MAX_PULSZEIT){
          zeit=0;
          status=0; //Timeout
        }else
        if (!durchfluss){ //Kein Durchfluss
          if (zeit >= MIN_PULSZEIT){
            status++; //Puls war kurz, aber nicht zu kurz
          }else{
            status=0; //Puls war zu kurz
          }
          zeit=0;
        }
        break;
        
      case 2:  //Teste:  1. kurze Pause 
        if (zeit > MAX_PAUSENZEIT){
          zeit=0;
          status++; 
        }else
        if (durchfluss){ //wieder Durchfluss
          if (zeit >= MIN_PAUSENZEIT){
            status++; //Pause war kurz, aber nicht zu kurz
          }else{
            status=0; //Pause war zu kurz
          }
          zeit=0;
        }
        break;	  
      
      case 3:  //Teste: 2. kurzer Puls 
        if (zeit > MAX_PULSZEIT){
          zeit=0;
          status=0; //Timeout
        }else
        if (!durchfluss){ //Kein Durchfluss
          zeit=0;
          status++; //Puls war kurz genug
        }
        break;

      case 4:  //Teste:  2. kurze Pause 
        if (zeit >= MAX_PAUSENZEIT){
          zeit=0;
          status++; 
        }else
        if (durchfluss){ //wieder Durchfluss
          if (zeit >=MIN_PAUSENZEIT){
            status++; //Pause war kurz, aber nicht zu kurz
          }else{
            status=0; //Pause war zu kurz
          }
          zeit=0;
        }
        break;	  
      
      case 5:  //Teste: 3. kurzer Puls 
        if (zeit >= MAX_PULSZEIT){
          zeit=0;
          status=0; //Timeout
        }else
        if (!durchfluss){ //Kein Durchfluss
          if (zeit >=MIN_PULSZEIT){
            status++; //Puls war kurz, aber nicht zu kurz
          }else{
            status=0; //Puls war zu kurz
          }
          zeit=0;
        }
        break;

      case 6:  //Teste:  lange Pause am Ende
        if (durchfluss){ //wieder Durchfluss - zu frueh!
          zeit=0;
          status=0; //Timeout
  	  }else
        if (zeit >= MIN_PAUSENZEIT_ENDE){
    	    //Magnetventil umschalten
          PORTB = PORTB ^ (1<<PB2);
          PORTB |=1; //Summer EIN
          zeit=0;
          status++; 
        }
        break;
      
      case 7: //Piepsen
        for (i=0; i<3700; i++){
          PORTB |=1; //Piezo High
          _delay_us(132);
          PORTB &=~1; //Piezo Low
          _delay_us(128);
        }
        zeit=0;
        status=0; //Von vorn
        break;

      default: 
        zeit=0;
  	  status=0; 
  	  break;  
    }
  }//While(1)
  return 0;
}
// EOF----------------------------------------------------------------------------------------------------------
